local math_max = math.max
local math_min = math.min
local math_lerp = math.lerp
local math_clamp = math.clamp

local mvec_distance = mvector3.distance
local mvec_set = mvector3.set

local gravity = 981

local launch_speed = 300 
local adjust_z = 50

local mass_look_up_modifier = 1
local projectile_mass = 0.5

function WeaponLaser:sample(curve_time, origin_position, origin_direction)
	local current_state = managers.player:get_current_state()
	local fire_position = origin_position or current_state:get_fire_weapon_position()
	local fire_direction = origin_direction or current_state:get_fire_weapon_direction()

	local velocity = fire_direction * launch_speed
	velocity.z = velocity.z + adjust_z

	local mass = math_max(mass_look_up_modifier * (1 + math_min(0, fire_direction.z)), 1)
	velocity = velocity * (mass / projectile_mass)

	local x = fire_position.x + (velocity.x * curve_time);
	local y = fire_position.y + (velocity.y * curve_time);
	local z = fire_position.z + (velocity.z * curve_time - (gravity * curve_time * curve_time / 2));

	return Vector3(x, y, z)
end

local temp_pos = Vector3()
function WeaponLaser:draw_laser_section(from, to, current_distance)
	local ray = self._unit:raycast("ray", from, to, "slot_mask", self._slotmask, self._ray_ignore_units and "ignore_unit" or nil, self._ray_ignore_units)

	if ray then
		if not self._is_npc then
			self._light:set_spot_angle_end(self._spot_angle_end)

			local distance = (current_distance + ray.distance)

			self._spot_angle_end = math_lerp(1, 18, distance / self._max_distance)

			self._light_glow:set_spot_angle_end(math_lerp(8, 80, distance / self._max_distance))

			local scale = (math_clamp(distance, self._max_distance - self._scale_distance, self._max_distance) - (self._max_distance - self._scale_distance)) / self._scale_distance
			scale = 1 - scale

			self._light:set_multiplier(scale)
			self._light_glow:set_multiplier(scale * 0.1)
		end

		self._brush:cylinder(ray.position, from, self._is_npc and 0.5 or 0.25)

		-- mvector3.set(temp_pos, mvec_l_dir)
		-- mvector3.multiply(temp_pos, 50)
		-- mvector3.negate(temp_pos)
		-- mvector3.add(temp_pos, ray.position)
		self._light:set_final_position(temp_pos)
		self._light_glow:set_final_position(temp_pos)

		return {true, distance}
	else
		self._light:set_final_position(to)
		self._light_glow:set_final_position(to)
		self._brush:cylinder(from, to, self._is_npc and 0.5 or 0.25)

		return {false, current_distance + mvec_distance(from, to)}
	end
end

local previous_position = Vector3()
local next_position = Vector3()

local sample_interval = 0.05
local samples = 100
function WeaponLaser:draw_laser_curve(origin_position, origin_direction)
	local current_distance = 0
	mvec_set(previous_position, origin_position)

	for i = 0, samples, 1 do
		local curve_time = i * sample_interval
		mvec_set(next_position, self:sample(curve_time, origin_position, origin_direction))

		local laser_section_attempt = self:draw_laser_section(previous_position, next_position, current_distance)
		local hit = laser_section_attempt[1]
		current_distance = laser_section_attempt[2]

		mvec_set(previous_position, next_position)

		if hit then
			break
		end
	end
end

local mvec_l_dir = Vector3()
local from = Vector3()

function WeaponLaser:update(unit, t, dt)
	local rotation = self._custom_rotation or self._laser_obj:rotation()
	mrotation.y(rotation, mvec_l_dir)

	if self._custom_position then
		mvector3.set(from, self._laser_obj:local_position())
		mvector3.rotate_with(from, rotation)
		mvector3.add(from, self._custom_position)
	else
		mvector3.set(from, self._laser_obj:position())
	end

	self:draw_laser_curve(from, mvec_l_dir)

	self._custom_position = nil
	self._custom_rotation = nil
end